% TrainFunc: train funct of the form TrainFunc = @trainRegLS
%       where trainRegLS is a function of the form
%           pars = trainRegLS(Xtr, Ytr, kern, sigma, lambda)
% TestFunc: test funct of the form TestFunc = @testRegLS
%       where testRegLS is a function of the form
%           yts = testRegLS(pars, Xts);
% func: the kernel function (or distance function) 
%       for example func='gaussian'
% X: the dataset (test set excluded)
% Y: the labels (test set excluded)
% perc: percentage of the dataset to be used for validation
% nrip: number of repetitions of the test for each couple of parameters
% intLambda: list of regularization parameters 
%       for example intLambda = [0.1, 0.05, 0.02, 0.01, 0.005, 0.002, 0.001, 0.0005, 0.0002, 0.0001]
%       use intLambda = [] if the algorithm doesn't have any regularization paramer
%       use intLambda = {[]} if the algorithm needs a dummy regularization
%       parameter
% intSigma: list of kernel parameters 
%       for example intSigma = [10, 5, 2, 1, 0.5, 0.2, 0.1, 0.05, 0.02, 0.01]
%       use intSigma = [] if the algorithm doesn't have any kernel paramer
%       use intSigma = {[]} if the algorithm needs a dummy kernel parameter
%       parameter
%
% Output:
% l, s: the couple of lambda and sigma that minimize the mean of the
%       validation error
% Vm, Vs: mean and variance of the validation error for each couple of parameters
% Tm, Ts: mean and variance of the error computed on the training set for each couple
%       of parameters
%
% Example of code:
% intLambda = 1:2:59;
% [X, Y] = MixGauss([[0;0],[1;1]],[0.5,0.25],1000);
% [l, s, Vm, Vs, Tm, Ts] = holdoutCV(@WrapperTrainKNN, @kNNClassify,'l2dist', X, Y, 0.5, 5, intLambda, {[]});
% plot(intLambda, Vm, 'b');
% hold on
% plot(intLambda, Tm, 'r');

function [l, s, Vm, Vs, Tm, Ts] = holdoutCV(TrainFunc, TestFunct, func, X, Y, perc, nrip, intLambda, intSigma)
    nSigma = numel(intSigma);
    nLambda = numel(intLambda);
    
    if nSigma == 0
        intSigma = [nan];
        nSigma = 1;
    elseif nLambda == 0
        intLambda = [nan];
        nLambda = 1;
    end
 
    
    n = size(X,1);
    ntr = ceil(n*(1-perc));
    
    Tm = zeros(nLambda, nSigma);
    Ts = zeros(nLambda, nSigma);
    Vm = zeros(nLambda, nSigma);
    Vs = zeros(nLambda, nSigma);
    
    ym = (max(Y) + min(Y))/2;
    
    il = 0;
    for l=intLambda
        il = il + 1;
        is = 0;
        for s=intSigma
            is = is + 1;
            for rip = 1:nrip
                I = randperm(n);
                Xtr = X(I(1:ntr),:);
                Ytr = Y(I(1:ntr),:);
                Xvl = X(I(ntr+1:end),:);
                Yvl = Y(I(ntr+1:end),:);
                
                pars = cell(1,nargout(TrainFunc));
                if isnumeric(l) && isnan(l)
                    [pars{:}] = TrainFunc(Xtr, Ytr, func, s);
                elseif isnumeric(s) && isnan(s)
                    [pars{:}] = TrainFunc(Xtr, Ytr, l); 
                else
                    [pars{:}] = TrainFunc(Xtr, Ytr, func, s, l);
                end
                
                
                
                trError =  calcErr(TestFunct(pars{:}, Xtr),Ytr, ym);
                Tm(il, is) = Tm(il, is) + trError;
                Ts(il, is) = Ts(il, is) + trError^2;
                
                valError  = calcErr(TestFunct(pars{:}, Xvl),Yvl, ym);
                Vm(il, is) = Vm(il, is) + valError;
                Vs(il, is) = Vs(il, is) + valError^2;
                
                str = sprintf('l\ts\tvalErr\ttrErr\n%f\t%f\t%f\t%f\n', l, s, valError, trError);
                disp(str);
            end
        end
    end
    
    Tm = Tm/nrip;
    Ts = Ts/nrip - Tm.^2;
    
    Vm = Vm/nrip;
    Vs = Vs/nrip - Vm.^2;
    
    [row, col] = find(Vm <= min(min(Vm)));
    
    l = intLambda(row(1));
    s = intSigma(col(1));
end

function err = calcErr(T, Y, m)
    vT = (T >= m);
    vY = (Y >= m);
    err = sum(vT ~= vY)/numel(Y);
end
